#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Implementation of Needleman Wunsch and derived algorithms
"""
import numpy as np
from collections import defaultdict
from scipy.spatial.distance import euclidean


class NWDistanceMatrix(object):
    """
    An object to hold the accumulated cost matrix for the
    Needleman Wunsch algorithm

    Parameters
    ----------
    gamma : float
        Gap parameter (for initializing the matrix)
    """
    def __init__(self, gamma = 0.1):
        self.gamma = float(gamma)
        self.val_dict = defaultdict(lambda: np.inf)
        self.indices_dict = dict()
        self.xdim = 0
        self.ydim = 0

    @property
    def nw_distance(self):
        """The total accumulated cost of alignment
        """
        return self[self.xdim, self.ydim]

    def __getitem__(self, indices):
        i, j = indices
        if i == 0 and j != 0:
            return self.gamma * j
        elif j == 0 and i != 0:
            return self.gamma * i
        elif i == j == 0:
            return 0
        else:
            return self.val_dict[i, j]

    def __setitem__(self, indices, values):
        if indices[0] > self.xdim:
            self.xdim = indices[0]
        if indices[1] > self.ydim:
            self.ydim = indices[1]
        self.val_dict[indices] = values[0]
        self.indices_dict[indices] = (values[1], values[2])

    def path_step(self, i, j):
        """
        Get the indices and type of message for backtracking
        """
        if i == j == 0:
            return (0, 0), 0
        elif i == 0 and j > 0:
            return (0, j - 1), 1
        elif j == 0 and i > 0:
            return (i - 1, 0), 2
        else:
            return self.indices_dict[i, j]

    @property
    def cost_matrix(self):
        """
        The alignment cost matrix
        """
        cost_matrix = np.ones((self.xdim + 1, self.ydim + 1)) * np.inf
        for i in range(self.xdim + 1):
            for j in range(self.ydim + 1):
                cost_matrix[i, j] = self[i, j]
        return cost_matrix


class NeedlemanWunsch(object):
    """
    Needleman-Wunsch algorithm for aligning sequences.
    """
    def __init__(self, 
                 metric = euclidean, 
                 gamma = 0.1):
        self.metric = metric
        self.gamma = gamma

    def __call__(self, X, Y, return_path=True,
                 window=None,
                 return_cost_matrix=False):
        X = X.astype(float)
        Y = Y.astype(float)
        len_X, len_Y = len(X), len(Y)
        if window is None:
            window = [(i, j) for i in range(len_X) for j in range(len_Y)]
        window = ((i + 1, j + 1) for i, j in window)

        nw_matrix = NWDistanceMatrix(self.gamma)

        for i, j in window:
            dt = self.metric(X[i-1], Y[j-1])

            nw_matrix[i, j] = min(
                (nw_matrix[i-1, j] + self.gamma, (i-1, j), 2),
                (nw_matrix[i, j-1] + self.gamma, (i, j-1), 1),
                (nw_matrix[i-1, j-1] + dt, (i-1, j-1), 0),
                key=lambda a: a[0]
            )

        nw_distance = nw_matrix[len_X, len_Y]

        out = (nw_distance, )

        if return_path:
            path = self.backtracking(nw_matrix)
            out += (path, )

        if return_cost_matrix:
            out += (nw_matrix, )

        return out

    def backtracking(self, nw_matrix):
        path = []
        i, j = nw_matrix.xdim, nw_matrix.ydim
        while not (i == j == 0):
            _, align = nw_matrix.path_step(i, j)
            if align == 0:
                path.append((i - 1, j - 1))
                i -= 1
                j -= 1
            elif align == 1:
                path.append((-1, j - 1))
                j -= 1
            elif align == 2:
                path.append((i - 1, -1))
                i -= 1
        path.reverse()
        return np.array(path, dtype=int)


# alias
NW = NeedlemanWunsch


class NeedlemanWunschDynamicTimeWarping(NeedlemanWunsch):
    """
    Needleman-Wunsch Dynamic Time Warping
    """
    def __init__(self, 
                 metric = euclidean, 
                 gamma = 0.1):
        super().__init__(metric=metric,
                         gamma=gamma)

    def __call__(self, X, Y, return_path=True,
                 window=None,
                 return_cost_matrix=False):

        X = X.astype(float)
        Y = Y.astype(float)
        len_X, len_Y = len(X), len(Y)
        if window is None:
            window = [(i, j) for i in range(len_X) for j in range(len_Y)]
        window = ((i + 1, j + 1) for i, j in window)

        nw_matrix = NWDistanceMatrix(self.gamma)

        pairwise_distance = defaultdict(lambda: float('inf'))

        pairwise_distance[0, 0] = 0

        for i, j in window:
            pairwise_distance[i, j] = self.metric(X[i-1], Y[j-1])

            nw_matrix[i, j] = min(
                (nw_matrix[i-1, j] + self.gamma, (i-1, j), 2),
                (nw_matrix[i, j-1] + self.gamma, (i, j-1), 1),
                (nw_matrix[i-1, j-1] + pairwise_distance[i, j], (i-1, j-1), 0),
                (nw_matrix[i-1, j-2] + pairwise_distance[i, j-1] +
                 pairwise_distance[i, j], (i, j-1), 0),
                (nw_matrix[i-2, j-1] + pairwise_distance[i-1, j] +
                 pairwise_distance[i, j], (i-1, j), 0),
                key=lambda a: a[0]
            )

        nw_distance = nw_matrix[len_X, len_Y]

        out = (nw_distance, )

        if return_path:
            path = self.backtracking(nw_matrix)
            out += (path, )

        if return_cost_matrix:
            out += (nw_matrix, )

        return out

    def backtracking(self, nw_matrix):
        path = []
        i, j = nw_matrix.xdim, nw_matrix.ydim

        while not (i == j == 0):
            ij, align = nw_matrix.path_step(i, j)
            if align == 0:
                path.append((i - 1, j - 1))
                i, j = ij
            elif align == 1:
                path.append((-1, j - 1))
                j -= 1
            elif align == 2:
                path.append((i - 1, -1))
                i -= 1
        path.reverse()
        return np.array(path, dtype=int)

# alias
NW_DTW = NeedlemanWunschDynamicTimeWarping



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains methods for note and sequence alignment.

sequence alignment:
- Dynamic Time Warping
- Needleman-Wunsch 

note alignment:
- Hidden Markov Model
- Greedy note alignment
- Path-augmented combinatorial alignment

most of the hidden markov machinery comes from 
https://github.com/neosatrapahereje/hiddenmarkov.git
this file contains specific variations and additions.

"""


from .dtw import DTW
from .nwtw import NW_DTW, NW
from .matchers import (AnchorPointNoteMatcher, 
                       AutomaticNoteMatcher)
from .utils import (node_array,
                    save_parangonada_csv)


#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains full note matcher classes.
"""
import numpy as np
from scipy.interpolate import interp1d

import time
from itertools import combinations
from scipy.special import binom

from .dtw import DTW
from .nwtw import NW_DTW, NW

from .preprocessors import (mend_note_alignments,
                            cut_note_arrays,
                            alignment_times_from_dtw)

################################### SYMBOLIC MATCHERS ###################################


class SimplestGreedyMatcher(object):
    """
    Create alignment in MAPS format (dict) by greedy pitch matching from performance and score note_array
    """
    def __call__(self, score_note_array, performance_note_array):
        alignment = []
        s_aligned = []
        p_aligned = []
        for s_note in score_note_array:
            sid = s_note['id']
            pid = None

            # filter performance notes with matching pitches
            matching_pitches = performance_note_array[s_note['pitch'] == performance_note_array['pitch']]

            for p_note in matching_pitches:
                # take first matching performance note that was not yet aligned
                if p_note not in p_aligned:
                    pid = str(p_note['id'])
                    p_aligned.append(p_note)
                    s_aligned.append(s_note)
                    break

            if pid is not None:
                alignment.append({'label': 'match', 'score_id': sid, 'performance_id': str(pid)})
            else:
                # if score note could not be aligned it counts as a deletion
                alignment.append({'label': 'deletion', 'score_id': sid})

        # check for unaligned performance notes (ie insertions)
        for p_note in performance_note_array:
            if p_note not in p_aligned:
                alignment.append({'label': 'insertion', 'performance_id': str(p_note['id'])})

        return alignment


class SequenceAugmentedGreedyMatcher(object):
    """
    Create alignment in MAPS format (dict) by sequence augmented pitch matching from performance and score note_array
    """
    def __init__(self):
        self.overlap = False

    def __call__(self, 
                 score_note_array, 
                 performance_note_array, 
                 alignment_times, 
                 shift=False, 
                 cap_combinations = 10000):
        alignment = []
        # s_aligned = []
        p_aligned = []

        # DTW gives non-unique times, sometimes...
        # TODO: safety net
        try:
            onset_time_conversion = interp1d(alignment_times[:, 0],
                                             alignment_times[:, 1],
                                             fill_value="extrapolate")
        except ValueError:
            if len(alignment_times) < 2:
                onset_time_conversion = \
                    lambda x: np.ones_like(x) * alignment_times[0, 1]  # noqa: E731

        score_pitches = np.unique(score_note_array["pitch"])
        # loop over pitches and align full sequences of matching pitches in correct order
        # if sequences mismatch in length, classify extra notes as insertions or deletions respectively
        for pitch in score_pitches:
         


            score_notes = score_note_array[pitch == score_note_array['pitch']]
            performance_notes = performance_note_array[pitch == performance_note_array['pitch']]
       
            score_notes_onsets = onset_time_conversion(score_notes["onset_beat"])
            score_notes_onsets_idx = np.argsort(score_notes_onsets)
            score_notes_onsets = score_notes_onsets[score_notes_onsets_idx]
    
            performance_notes_onsets = performance_notes["onset_sec"]
            performance_notes_onsets_idx = np.argsort(performance_notes_onsets)
            performance_notes_onsets = performance_notes_onsets[performance_notes_onsets_idx]

            score_no = score_notes_onsets.shape[0]
            performance_no = performance_notes_onsets.shape[0]
    
            common_no = min(score_no, performance_no)
            extra_no = max(score_no, performance_no)-common_no
            score_longer = np.argmax([performance_no, score_no])

            if score_longer:
                longt = score_notes_onsets
                shortt = performance_notes_onsets
                longid = score_notes_onsets_idx
                shortid = performance_notes_onsets_idx
            else:
                longt = performance_notes_onsets
                shortt = score_notes_onsets
                longid = performance_notes_onsets_idx
                shortid = score_notes_onsets_idx

            diffs = dict()
 
            if shift:
                if cap_combinations is not None:
                    combination_number = binom(max(score_no, performance_no), extra_no)
                    if combination_number > cap_combinations:
                        combs = [np.random.choice(max(score_no, performance_no), extra_no, replace=False) for n in range(cap_combinations)]  

                        print("high number of combinations: ", combination_number, "low number sampled ", len(combs))
                    else:
                        combs = combinations(range(max(score_no, performance_no)), extra_no)
                else:
                    combs = combinations(range(max(score_no, performance_no)), extra_no)
                for omit_idx in combinations(range(max(score_no, performance_no)), extra_no):
                    shortenedt = np.delete(longt,list(omit_idx))
                    optimal_shift = np.mean(shortenedt-shortt)
                    shift_diff = np.sum(np.abs(shortenedt-shortt-optimal_shift*np.ones_like(shortenedt))**2)
                    diffs[shift_diff] = list(omit_idx)


            else:
                
                if cap_combinations is not None:
                    combination_number = binom(max(score_no, performance_no), extra_no)
                    if combination_number > cap_combinations:
                        combs = [np.random.choice(max(score_no, performance_no), extra_no, replace=False) for n in range(cap_combinations)]  

                        print("high number of combinations: ", combination_number, "low number sampled ", len(combs))
                    else:
                        combs = combinations(range(max(score_no, performance_no)), extra_no)
                else:
                    combs = combinations(range(max(score_no, performance_no)), extra_no)
                for omit_idx in combs:

                    shortenedt = np.delete(longt,list(omit_idx))
                    diff = np.sum(np.abs(shortenedt-shortt)**2)
                    diffs[diff] = list(omit_idx)

            best_omit_dist = np.min(list(diffs.keys()))
            best_omit_idx = diffs[best_omit_dist]

            # get the arrays of actual onset times
            aligns = np.delete(longt, best_omit_idx)
            nonaligns = longt[best_omit_idx]
            # get the idx of the original note_arrays
            align_ids = np.delete(longid, best_omit_idx)
            nonalign_ids = longid[best_omit_idx]

            if score_longer:
                for sid, pid in zip(score_notes["id"][align_ids], performance_notes["id"][performance_notes_onsets_idx]):
                    alignment.append({'label': 'match', 'score_id': sid, 'performance_id': str(pid)})
                    p_aligned.append(str(pid))

                for sid in score_notes["id"][nonalign_ids]:
                    alignment.append({'label': 'deletion', 'score_id': sid})

            else:
                for sid, pid in zip(score_notes["id"][score_notes_onsets_idx], performance_notes["id"][align_ids]):
                    alignment.append({'label': 'match', 'score_id': sid, 'performance_id': str(pid)})
                    p_aligned.append(str(pid))

                for pid in performance_notes["id"][nonalign_ids]:
                    alignment.append({'label': 'insertion', 'performance_id': str(pid)})
                    p_aligned.append(str(pid))

        # check for unaligned performance notes (ie insertions)
        for p_note in performance_note_array:
            if str(p_note['id']) not in p_aligned:
                alignment.append({'label': 'insertion', 'performance_id': str(p_note['id'])})

        return alignment
    
    
################################### FULL MODEL MATCHERS ###################################


class PianoRollSequentialMatcher(object):
    def __init__(self,
                 note_matcher=DTW,
                 matcher_kwargs=dict(metric="euclidean"),
                 node_cutter=cut_note_arrays,
                 node_mender=mend_note_alignments,
                 symbolic_note_matcher=SequenceAugmentedGreedyMatcher(),
                 greedy_symbolic_note_matcher=SimplestGreedyMatcher(),
                 alignment_type="dtw",
                 SCORE_FINE_NODE_LENGTH=0.25,
                 s_time_div=16,
                 p_time_div=16,
                 sfuzziness=0.5,
                 pfuzziness=0.5,
                 window_size=1,
                 pfuzziness_relative_to_tempo=True,
                 shift_onsets=False,
                 cap_combinations=None):

        self.note_matcher = note_matcher(**matcher_kwargs)
        self.symbolic_note_matcher = symbolic_note_matcher
        self.node_cutter = node_cutter
        self.node_mender = node_mender
        self.greedy_symbolic_note_matcher = greedy_symbolic_note_matcher
        self.alignment_type = alignment_type
        self.SCORE_FINE_NODE_LENGTH = SCORE_FINE_NODE_LENGTH
        self.s_time_div = s_time_div
        self.p_time_div = p_time_div
        self.sfuzziness = sfuzziness
        self.pfuzziness = pfuzziness
        self.window_size = window_size
        self.pfuzziness_relative_to_tempo = pfuzziness_relative_to_tempo
        self.shift_onsets = shift_onsets
        self.cap_combinations = cap_combinations

    def __call__(self, score_note_array,
                 performance_note_array, 
                 alignment_times):
        
        # cut arrays to windows
        score_note_arrays, performance_note_arrays = self.node_cutter(
            performance_note_array,
            score_note_array,
            np.array(alignment_times),
            sfuzziness=self.sfuzziness, 
            pfuzziness=self.pfuzziness,
            window_size=self.window_size,
            pfuzziness_relative_to_tempo=self.pfuzziness_relative_to_tempo)

        # compute windowed alignments
        note_alignments = []
        dtw_al = []

        for window_id in range(len(score_note_arrays)):  
            if self.alignment_type == "greedy":
                alignment = self.greedy_symbolic_note_matcher(
                    score_note_arrays[window_id],
                    performance_note_arrays[window_id])
                note_alignments.append(alignment)
            else:
                # _____________ fine alignment ____________
                if self.alignment_type == "dtw":
                    if score_note_arrays[window_id].shape[0] == 0 or \
                        performance_note_arrays[window_id].shape[0] == 0:
                        # for empty arrays fall back to linear
                        dtw_alignment_times = np.array(alignment_times)[
                                window_id:window_id+2, :]

                    else:    
                        dtw_alignment_times = alignment_times_from_dtw(
                            score_note_arrays[window_id],
                            performance_note_arrays[window_id],
                            matcher=self.note_matcher,
                            SCORE_FINE_NODE_LENGTH=self.SCORE_FINE_NODE_LENGTH,
                            s_time_div=self.s_time_div,
                            p_time_div=self.p_time_div)
                else:
                    dtw_alignment_times = np.array(alignment_times)[
                        window_id:window_id+2, :]

                dtw_al.append(dtw_alignment_times)
               
                # distance augmented greedy align
                fine_local_alignment = self.symbolic_note_matcher(
                    score_note_arrays[window_id],
                    performance_note_arrays[window_id],
                    dtw_alignment_times,
                    shift=self.shift_onsets,
                    cap_combinations=self.cap_combinations)

                note_alignments.append(fine_local_alignment)

                

        # MEND windows to global alignment
        global_alignment, score_alignment, \
            performance_alignment = self.node_mender(note_alignments, 
                                                    performance_note_array,
                                                    score_note_array, 
                                                    node_times=np.array(alignment_times),
                                                    symbolic_note_matcher= self.symbolic_note_matcher,
                                                    max_traversal_depth=1500)

        return global_alignment


class PianoRollNoNodeMatcher(object):
    def __init__(self,
                 note_matcher=DTW,
                 matcher_kwargs=dict(metric="euclidean"),
                 node_cutter=cut_note_arrays,
                 node_mender=mend_note_alignments,
                 symbolic_note_matcher=SequenceAugmentedGreedyMatcher(),
                 greedy_symbolic_note_matcher=SimplestGreedyMatcher(),
                 alignment_type="dtw",
                 SCORE_FINE_NODE_LENGTH=0.25,
                 s_time_div=16,
                 p_time_div=16,
                 sfuzziness=0.5,
                 pfuzziness=0.5,
                 window_size=1,
                 pfuzziness_relative_to_tempo=True,
                 shift_onsets=False,
                 cap_combinations=None):

        self.note_matcher = note_matcher(**matcher_kwargs)
        self.symbolic_note_matcher = symbolic_note_matcher
        self.node_cutter = node_cutter
        self.node_mender = node_mender
        self.greedy_symbolic_note_matcher = greedy_symbolic_note_matcher
        self.alignment_type = alignment_type
        self.SCORE_FINE_NODE_LENGTH = SCORE_FINE_NODE_LENGTH
        self.s_time_div = s_time_div
        self.p_time_div = p_time_div
        self.sfuzziness = sfuzziness
        self.pfuzziness = pfuzziness
        self.window_size = window_size
        self.pfuzziness_relative_to_tempo = pfuzziness_relative_to_tempo
        self.shift_onsets = shift_onsets
        self.cap_combinations = cap_combinations

    def __call__(self, score_note_array,
                 performance_note_array,
                 verbose_time=False):
        
        t1 = time.time()
        # start with DTW
        dtw_alignment_times_init = alignment_times_from_dtw(
                            score_note_array,
                            performance_note_array,
                            matcher=self.note_matcher,
                            SCORE_FINE_NODE_LENGTH=4.0,
                            s_time_div=self.s_time_div,
                            p_time_div=self.p_time_div
                            )
        # cut arrays to windows
        t11 = time.time()
        if verbose_time:
            print(format(t11-t1, ".3f"), "sec : Initial coarse DTW pass")
        score_note_arrays, performance_note_arrays = self.node_cutter(
            performance_note_array,
            score_note_array,
            np.array(dtw_alignment_times_init),
            sfuzziness=self.sfuzziness, 
            pfuzziness=self.pfuzziness,
            window_size=self.window_size,
            pfuzziness_relative_to_tempo=self.pfuzziness_relative_to_tempo)

        # compute windowed alignments
        note_alignments = []
        dtw_al = []

        t2 = time.time()
        if verbose_time:
            print(format(t2-t11, ".3f"), "sec : Cutting")
            
        for window_id in range(len(score_note_arrays)):
            if self.alignment_type == "greedy":
                alignment = self.greedy_symbolic_note_matcher(
                    score_note_arrays[window_id],
                    performance_note_arrays[window_id])
                note_alignments.append(alignment)
            else:
                # _____________ fine alignment ____________
                if self.alignment_type == "dtw":
                    if score_note_arrays[window_id].shape[0] == 0 or \
                        performance_note_arrays[window_id].shape[0] == 0:
                        # for empty arrays fall back to linear
                        dtw_alignment_times = np.array(dtw_alignment_times_init)[
                                window_id:window_id+2, :]

                    else:    
                        dtw_alignment_times = alignment_times_from_dtw(
                            score_note_arrays[window_id],
                            performance_note_arrays[window_id],
                            matcher=self.note_matcher,
                            SCORE_FINE_NODE_LENGTH=self.SCORE_FINE_NODE_LENGTH,
                            s_time_div=self.s_time_div,
                            p_time_div=self.p_time_div)
                else:
                    dtw_alignment_times = np.array(dtw_alignment_times_init)[
                        window_id:window_id+2, :]

                dtw_al.append(dtw_alignment_times)
                
                # distance augmented greedy align
                fine_local_alignment = self.symbolic_note_matcher(
                    score_note_arrays[window_id],
                    performance_note_arrays[window_id],
                    dtw_alignment_times,
                    shift=self.shift_onsets,
                    cap_combinations=self.cap_combinations)

                note_alignments.append(fine_local_alignment)
        t41 = time.time()
        if verbose_time:
            print(format(t41-t2, ".3f"), "sec : Fine-grained DTW passes, symbolic matching")

        
        # MEND windows to global alignment
        global_alignment, score_alignment, \
            performance_alignment = self.node_mender(note_alignments, 
                                                    performance_note_array,
                                                    score_note_array, 
                                                    node_times=np.array(dtw_alignment_times_init),
                                                    symbolic_note_matcher= self.symbolic_note_matcher,
                                                    max_traversal_depth=150)
        t5 = time.time()
        if verbose_time:
            print(format(t5-t41, ".3f"), "sec : Mending")

        return global_alignment

# alias
AutomaticNoteMatcher = PianoRollNoNodeMatcher

# alias
AnchorPointNoteMatcher = PianoRollSequentialMatcher



from partitura.utils.music import (ensure_notearray)

from partitura.musicanalysis.performance_codec import ( 
                                    to_matched_score,
                                    get_unique_onset_idxs
                                    )

import numpy as np
import os
from scipy.interpolate import interp1d

################################### PARANGONADA EXPORT ###################################


def alignment_dicts_to_array(alignment):
    """
    create structured array from list of dicts type alignment.

    Parameters
    ----------
    alignment : list
        A list of note alignment dictionaries.

    Returns
    -------
    alignarray : structured ndarray
        Structured array containing note alignment.
    """
    fields = [
        ("idx", "i4"),
        ("matchtype", "U256"),
        ("partid", "U256"),
        ("ppartid", "U256"),
    ]

    array = []
    # for all dicts create an appropriate entry in an array:
    # match = 0, deletion  = 1, insertion = 2
    for no, i in enumerate(alignment):

        if i["label"] == "match":
            array.append((no, "0", i["score_id"], str(i["performance_id"])))
        elif i["label"] == "insertion":
            array.append((no, "2", "undefined", str(i["performance_id"])))
        elif i["label"] == "deletion":
            array.append((no, "1", i["score_id"], "undefined"))

    alignarray = np.array(array, dtype=fields)

    return alignarray


def save_parangonada_csv(
    alignment,
    performance_data,
    score_data,
    outdir = None,
    zalign = None,
    feature = None
):
    """
    Save an alignment for visualization with parangonda.

    Parameters
    ----------
    alignment : list
        A list of note alignment dictionaries.
    performance_data : Performance, PerformedPart, structured ndarray
        The performance information
    score_data : ScoreLike
        The musical score. A :class:`partitura.score.Score` object,
        a :class:`partitura.score.Part`, a :class:`partitura.score.PartGroup` or
        a list of these.
    outdir : PathLike
        A directory to save the files into.
    ppart : PerformedPart, structured ndarray
        A PerformedPart or its note_array.
    zalign : list, optional
        A second list of note alignment dictionaries.
    feature : list, optional
        A list of expressive feature dictionaries.

    Returns
    -------
    perf_note_array : np.ndarray
        The performance note array. Only returned if `outdir` is None.
    score_note_array: np.ndarray
        The note array from the score. Only returned if `outdir` is None.
    alignarray: np.ndarray
    zalignarray: np.ndarray
    featurearray: np.ndarray
    """

    score_note_array = ensure_notearray(score_data)

    perf_note_array = ensure_notearray(performance_data)

    ffields = [
        ("velocity", "<f4"),
        ("timing", "<f4"),
        ("articulation", "<f4"),
        ("id", "U256"),
    ]

    farray = []
    notes = list(score_note_array["id"])
    if feature is not None:
        # veloctiy, timing, articulation, note
        for no, i in enumerate(list(feature["id"])):
            farray.append(
                (
                    feature["velocity"][no],
                    feature["timing"][no],
                    feature["articulation"][no],
                    i,
                )
            )
    else:
        for no, i in enumerate(notes):
            farray.append((0, 0, 0, i))

    featurearray = np.array(farray, dtype=ffields)
    alignarray = alignment_dicts_to_array(alignment)

    if zalign is not None:
        zalignarray = alignment_dicts_to_array(zalign)
    else:  # if no zalign is available, save the same alignment twice
        zalignarray = alignment_dicts_to_array(alignment)

    if outdir is not None:
        np.savetxt(
            os.path.join(outdir, "ppart.csv"),
            # outdir + os.path.sep + "perf_note_array.csv",
            perf_note_array[
                [
                    "onset_sec",
                    "duration_sec",
                    "pitch",
                    "velocity",
                    "track",
                    "channel",
                    "id",
                ]
            ],
            fmt="%.20s",
            delimiter=",",
            header=",".join(
                [
                    "onset_sec",
                    "duration_sec",
                    "pitch",
                    "velocity",
                    "track",
                    "channel",
                    "id",
                ]
            ),
            comments="",
        )
        np.savetxt(
            os.path.join(outdir, "part.csv"),
            # outdir + os.path.sep + "score_note_array.csv",
            score_note_array,
            fmt="%.20s",
            delimiter=",",
            header=",".join(score_note_array.dtype.names),
            comments="",
        )
        np.savetxt(
            os.path.join(outdir, "align.csv"),
            # outdir + os.path.sep + "align.csv",
            alignarray,
            fmt="%.20s",
            delimiter=",",
            header=",".join(alignarray.dtype.names),
            comments="",
        )
        np.savetxt(
            os.path.join(outdir, "zalign.csv"),
            # outdir + os.path.sep + "zalign.csv",
            zalignarray,
            fmt="%.20s",
            delimiter=",",
            header=",".join(zalignarray.dtype.names),
            comments="",
        )
        np.savetxt(
            os.path.join(outdir, "feature.csv"),
            # outdir + os.path.sep + "feature.csv",
            featurearray,
            fmt="%.20s",
            delimiter=",",
            header=",".join(featurearray.dtype.names),
            comments="",
        )
    else:
        return (
            perf_note_array,
            score_note_array,
            alignarray,
            zalignarray,
            featurearray,
        )


################################### ANCHOR POINT GENERAITON ###################################


def node_array(part, 
                ppart,
                alignment,
                tapping_noise=False,
                node_interval=1, 
                start_beat=None,
                nodes_in_beats=True):
    """
    generates an array of nodes, corresponding score time point and 
    performance time point tuples, spacing given by note_interval.
    
    Args:
        part (partitura.Part): a part object
        ppart (partitura.PerformedPart): a performedpart object
        alignment (List(Dict)): an alignment

    Returns:
        np.ndarray: a minimal, aligned 
            node note array.

    """

    ppart.sustain_pedal_threshold = 128

    matched_score, snote_ids = to_matched_score(part, ppart, alignment)

    # get unique onsets
    nid_dict = dict((n.id, i) for i, n in enumerate(part.notes_tied))
    matched_subset_idxs = np.array([nid_dict[nid] for nid in snote_ids])

    part_note_array = ensure_notearray(part,
                                        include_time_signature=True)

    score_onsets = part_note_array[matched_subset_idxs]['onset_beat']
    # changed from onset_beat
    unique_onset_idxs = get_unique_onset_idxs(score_onsets)

    smatched_score_onset = notewise_to_onsetwise(matched_score['onset'],
                                                    unique_onset_idxs,
                                                    aggregation_func=np.mean)
    pmatched_score_onset = notewise_to_onsetwise(matched_score['p_onset'],
                                                    unique_onset_idxs,
                                                    aggregation_func=np.mean)
    
    if nodes_in_beats:
        beat_times, sbeat_times = beat_times_from_matched_score(
            smatched_score_onset, 
            pmatched_score_onset,
            tapping_noise=tapping_noise,
            node_interval=node_interval,
            start_beat=start_beat)
    else: 
        beat_times, sbeat_times = measure_times_from_matched_score(
            smatched_score_onset, 
            pmatched_score_onset, 
            part, 
            tapping_noise=tapping_noise,
            measure_interval=node_interval,
            start_measure=start_beat)

    alignment_times = [*zip(sbeat_times, beat_times)]

    return alignment_times


def beat_times_from_matched_score(x, y,
                                  tapping_noise=False,
                                  node_interval=1.0,
                                  start_beat=None):

    beat_times_func = interp1d(x, y, kind="linear", fill_value="extrapolate")

    min_beat = np.ceil(x.min())  # -node_interval
    min_beat_original = min_beat
    max_beat = np.floor(x.max())  # +node_interval

    if start_beat is not None:
        print("first node at ", start_beat, " first beat at ", min_beat)
        min_beat = start_beat
        
    sbeat_times = np.arange(min_beat, max_beat, node_interval)
    # prepend and append very low and high values to make sure every beat/beat annotation falls between two sbeat_times/beat_times
    sbeat_times = np.append(np.append(min_beat_original-max(100, node_interval),sbeat_times),max_beat+max(100, node_interval))
    beat_times = beat_times_func(sbeat_times)

    if tapping_noise:
        beat_times += np.random.normal(0.0, tapping_noise,
                                       beat_times.shape)

    return beat_times, sbeat_times


def measure_times_from_matched_score(x, y, part, 
                                  tapping_noise=False,
                                  measure_interval=1,
                                  start_measure=None):

    beat_times_func = interp1d(x, y, kind="linear", fill_value="extrapolate")

    measure_times_in_part = [part.beat_map(measure.start.t) for measure in part.iter_all(partitura.score.Measure)]
    if start_measure is None:
        start_measure=0
    
    smeasure_idx = np.arange(start_measure, len(measure_times_in_part), measure_interval, dtype = int)
    smeasure_times = np.array(measure_times_in_part)[smeasure_idx]
    # prepend and append very low and high values to make sure every beat/beat annotation falls between two sbeat_times/beat_times
    smeasure_times = np.append(np.append(smeasure_times[0]-max(100, measure_interval),smeasure_times),smeasure_times[-1]+max(100, measure_interval))
    pmeasure_times = beat_times_func(smeasure_times)

    if tapping_noise:
        pmeasure_times += np.random.normal(0.0, tapping_noise, pmeasure_times.shape)

    return pmeasure_times, smeasure_times


def notewise_to_onsetwise(notewise_inputs, 
                          unique_onset_idxs, 
                          aggregation_func=np.mean):
    """Agregate onset times per score onset
    """
    
    onsetwise_inputs = np.zeros(len(unique_onset_idxs),
                                dtype=notewise_inputs.dtype)

    for i, uix in enumerate(unique_onset_idxs):
        onsetwise_inputs[i] = aggregation_func(notewise_inputs[uix])
    return onsetwise_inputs


def expand_grace_notes(note_array, backwards_time=0.2):
    """
    expand the duration of gracenotes in a note_array and reset their onset by a timespan called backwards_time
    """
    grace_note_onsets = np.unique(
        note_array["onset_beat"][note_array["duration_beat"] == 0.0])
    for gno in grace_note_onsets:
        mask = np.all([note_array["duration_beat"] == 0.0,
                       note_array["onset_beat"] == gno],
                      axis=0)
        number_of_grace_notes = mask.sum()
        note_array["onset_beat"][mask] -= np.linspace(backwards_time,0.0, number_of_grace_notes+1)[:-1]
        note_array["duration_beat"][mask] += backwards_time/(number_of_grace_notes)
    return note_array


def convert_grace_to_insertions(alignment):
    """
    relabel all ornament alignments as insertions
    """
    new_alignment = [al for al in alignment if al["label"] != "ornament"]
    new_alignment_o = [al for al in alignment if al["label"] == "ornament"]
    for al in new_alignment_o:
        new_al = {'label': 'insertion', 'performance_id': al["performance_id"]}
        new_alignment.append(new_al)
    return new_alignment




#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains dynamic time warping methods.
"""

import numpy as np
from scipy.spatial.distance import cdist

def element_of_metric(vec1, vec2):
    """
    metric that evaluates occurence of vec2 (scalar) in vec1 (vector n-dim)
    """
    return 1 - np.sum(vec2 == vec1)


def l2(vec1, vec2):
    """
    l2 metric between vec1 and vec2
    """
    return np.sqrt(np.sum((vec2 - vec1)**2))


class DynamicTimeWarping(object):
    """
    pure python vanilla Dynamic Time Warping
    """
    
    def __init__(self, 
                 metric= 'euclidean'):
        self.metric = metric

    def __call__(self, X, Y, return_path=True,
                 return_cost_matrix=False):

        X = np.asanyarray(X, dtype=float)
        Y = np.asanyarray(Y, dtype=float)
        # Compute pairwise distance
        D = cdist(X, Y, self.metric)
        # Compute accumulated cost matrix
        dtwd_matrix = dtw_dmatrix_from_pairwise_dmatrix(D)
        dtwd_distance = dtwd_matrix[-1, -1]

        # Output
        out = (dtwd_distance, )

        if return_path:
            # Compute alignment path
            path = dtw_backtracking(dtwd_matrix)
            out += (path,)
        if return_cost_matrix:
            out += (dtwd_matrix, )
        return out

# alias
DTW = DynamicTimeWarping
    
def dtw_backtracking(dtwd):
    """
    Decode path from the accumulated dtw cost matrix.

    Parameters
    ----------
    dtwd : np.ndarray
        Accumulated cost matrix (computed with 
        `dtw_dmatrix_from_pairwise_dmatrix`)
    
    Returns
    -------
    path : np.ndarray
       A 2D array of size (n_steps, 2), where i-th row has elements 
       (i_m, i_n) where i_m represents the index in the input array
       and i_n represents the corresponding index in the reference array.
    """
    
    N = dtwd.shape[0]
    M = dtwd.shape[1]

    n = N - 1
    m = M - 1

    step = [n, m]

    path = [step]

    # Initialize step choices
    choices = np.zeros((3, 2), dtype=int)
    # Initialize a vector for candidate distances
    dtwd_candidates = np.zeros(3, dtype=float)
    # initialize boolean variables for stopping decoding
    crit = True

    while crit:

        if n == 0:
            # next point in the path
            m = m - 1

        elif m == 0:
            # next point in the path
            n = n - 1

        else:
            # step sizes
            choices[0, 0] = n - 1
            choices[0, 1] = m - 1
            choices[1, 0] = n - 1
            choices[1, 1] = m
            choices[2, 0] = n
            choices[2, 1] = m - 1

            # accumulated distance from the previous step
            # to the next
            dtwd_candidates[0] = dtwd[n - 1, m - 1]
            dtwd_candidates[1] = dtwd[n - 1, m]
            dtwd_candidates[2] = dtwd[n, m - 1]

            # select the best candidate
            p_l_i = np.argmin(dtwd_candidates)

            # update next indices
            n = choices[p_l_i, 0]
            m = choices[p_l_i, 1]

        step = [n, m]
        # append next step to the path
        path.append(step)

        if n == 0 and m == 0:
            crit = False

    return np.array(path[::-1], dtype=int)


# def cdist(arr1, arr2, metric):
#     """
#     compute array of pairwise distances between 
#     the elements of two arrays given a metric
    
#     Parameters
#     ----------
#     arr1: numpy nd array
    
#     arr2: numpy nd array
    
#     metric> callable
#         a metric function
    
#     Returns
#     -------
    
#     pdist_array: numpy 2d array
#         array of pairwise distances
#     """
#     pdist_array = np.ones((arr1.shape[0],arr2.shape[0]))*np.inf
#     for i in range(arr1.shape[0]):
#         for j in range(arr2.shape[0]):
#             pdist_array[i, j] = metric(arr1[i], arr2[j])
#     return pdist_array


def dtw_dmatrix_from_pairwise_dmatrix(D):
    """
    compute dynamic time warping cost matrix 
    from a pairwise distance matrix

    Parameters
    ----------
    D : double array
        Pairwise distance matrix (computed e.g., with `cdist`).

    Returns
    -------
    dtwd : np.ndarray
        Accumulated cost matrix
    """
    # Initialize arrays and helper variables
    M = D.shape[0]
    N = D.shape[1]
    # the dtwd distance matrix is initialized with INFINITY
    dtwd = np.ones((M + 1, N + 1),dtype=float) * np.inf
    
    # Compute the distance iteratively
    dtwd[0, 0] = 0
    for i in range(1, M + 1):
        for j in range(1, N + 1):
            c = D[i - 1, j - 1]
            insertion = dtwd[i - 1, j]
            deletion = dtwd[i, j - 1]
            match = dtwd[i - 1, j - 1]
            dtwd[i, j] = c + min((insertion, deletion, match))

    return (dtwd[1:, 1:])



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
The top level of the package contains functions to
note-align symbolic music data.
"""
import pkg_resources
EXAMPLE = pkg_resources.resource_filename("parangonar", 
                                          "assets/mozart_k265_var1.match")


from .match import AnchorPointNoteMatcher, AutomaticNoteMatcher
from .evaluate import fscore_alignments, plot_alignment

__all__ = [
    "AnchorPointNoteMatcher",
    "AutomaticNoteMatcher",
    "fscore_alignments",
]



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains preprocessing methods
"""

import numpy as np
from scipy.interpolate import interp1d

from partitura.utils.music import (compute_pianoroll)

from .dtw import DTW
from .nwtw import NW_DTW, NW


################################### HELPERS ###################################


def alignment_times_from_dtw(score_note_array, 
                             performance_note_array,
                             matcher=DTW(),
                             SCORE_FINE_NODE_LENGTH=1.0,
                             s_time_div=16, p_time_div=16):
    """
    
    Coarse time warping to generate anchor points
    Returns an array, each row is tuple of corresponding 
    times in the score array and the perfromance array,
    respectively.

    Args:
        score_note_array (_type_): _description_
        performance_note_array (_type_): _description_
        matcher (_type_, optional): _description_. Defaults to DTW().
        SCORE_FINE_NODE_LENGTH (float, optional): _description_. Defaults to 1.0.
        s_time_div (int, optional): _description_. Defaults to 16.
        p_time_div (int, optional): _description_. Defaults to 16.

    Returns:
        _type_: _description_
    """
    # _____________ fine alignment ____________
    # compute proper piano rolls
    s_pianoroll = compute_pianoroll(score_note_array,
                                    time_div=s_time_div,
                                    remove_drums=False).toarray()
    p_pianoroll = compute_pianoroll(performance_note_array,
                                    time_div=p_time_div,
                                    remove_drums=False).toarray()
    # make piano rolls binary
    p_pianoroll_ones = np.zeros_like(p_pianoroll)
    p_pianoroll_ones[p_pianoroll > 0.0] = 1.0
    # align the piano rolls
    _, path = matcher(s_pianoroll.T, p_pianoroll_ones.T)
    # compute an alignment of times using the DTW path
    path_array = np.array(path)

    # Post-process path for NW-DTW (insertion and deletions
    # are denoted as -1 in the path)
    if isinstance(matcher, (NW_DTW, NW)):
        valid_idxs = np.where(
            np.logical_and(path_array[:, 0] != -1,
                           path_array[:, 1] != -1))[0]
        path_array = path_array[valid_idxs]

        if len(path_array) == 0:
            path_array = np.array([(path[:, 0].min(), path[:, 1].min()),
                                   (path[:, 1].max(), path[:, 1].max())],
                                  dtype=int)

    times_score = path_array[:, 0] / s_time_div
    times_performance = path_array[:, 1] / p_time_div

    u_times_score = np.unique(times_score)
    u_times_score_idxs = [np.where(times_score == u)[0]
                          for u in u_times_score]
    u_times_performance = np.array(
        [np.min(times_performance[ui])
         for ui in u_times_score_idxs])

    try:
        # Use a mapping to deal with missing values (due to
        # insertions and deletions in NW-related methods)
        # It should not affect the behavior of DTW methods
        # CC: I will check this just in case ;)
        stime_to_ptime_map = interp1d(x=u_times_score,
                                      y=u_times_performance,
                                      kind='linear',
                                      bounds_error=False,
                                      fill_value='extrapolate')
    except ValueError:
        # if there is only one element in u_times_score
        stime_to_ptime_map = \
            lambda x: np.ones_like(x) * u_times_performance[0]  # noqa: E731

    min_score = times_score.min()
    max_score = max(score_note_array["onset_beat"].max() -
                    score_note_array["onset_beat"].min(),
                    SCORE_FINE_NODE_LENGTH)
    max_performance = max(performance_note_array["onset_sec"].max() -
                          performance_note_array["onset_sec"].min(),
                          SCORE_FINE_NODE_LENGTH)

    cut_times_score = np.r_[
        np.arange(min_score, max_score,
                  SCORE_FINE_NODE_LENGTH),
        max_score]

    cut_times_performance = np.clip(
        stime_to_ptime_map(cut_times_score),
        a_min=0,
        a_max=max_performance)

    cut_times_performance += performance_note_array["onset_sec"].min()
    cut_times_score += score_note_array["onset_beat"].min()

    dtw_alignment_times = np.column_stack((cut_times_score,
                                           cut_times_performance))
    return dtw_alignment_times


def traverse_the_alignment_graph(start_id, 
                                 score_ids, 
                                 performance_ids, 
                                 performance_alignment, 
                                 score_alignment, counter, 
                                 max_depth=150):
    #score_ids = [input_id]
    #performance_id = []
    if start_id not in score_ids and counter < max_depth:
        #print("appendin", start_id)
        score_ids.append(start_id)
        new_pids = score_alignment[start_id]
        for pid in new_pids:
            #print("loop point", pid)
            if pid not in performance_ids:
                #print("appendin perf", pid)
                performance_ids.append(pid)
                new_sids = performance_alignment[pid]
                for sid in new_sids:
                    counter += 1
                    traverse_the_alignment_graph(sid, score_ids, performance_ids, performance_alignment, score_alignment, counter, max_depth)
                    
            else:
                continue
    elif counter == 150:
        print("max recursion depth in note graph")
        pass
    else:
        pass
        #print("done")

################################### SEGMENT CUTTING ###################################


def cut_note_arrays(performance_note_array, 
                    score_note_array, 
                    alignment,
                    sfuzziness=0.0, 
                    pfuzziness=0.0, 
                    window_size=1,
                    pfuzziness_relative_to_tempo=False):
    """
    cut note arrays into two lists of corresponding
    note array segments based on anchor points given 
    as "alignment": n by 2 array, first column score, 
    second performace times.


    Args:
        performance_note_array (_type_): _description_
        score_note_array (_type_): _description_
        alignment (_type_): _description_
        sfuzziness (float, optional): _description_. Defaults to 0.0.
        pfuzziness (float, optional): _description_. Defaults to 0.0.
        window_size (int, optional): _description_. Defaults to 1.
        pfuzziness_relative_to_tempo (bool, optional): _description_. Defaults to False.

    Returns:
        _type_: _description_
    """
    
    score_note_arrays = []
    performance_note_arrays = []

    if not pfuzziness_relative_to_tempo:
        local_pfuzzines = np.ones_like(alignment[:,1])*pfuzziness
    else:
        approximate_tempo = interp1d(alignment[:-1,0],
                                     np.diff(alignment[:,1])/np.diff(alignment[:,0]),
                                     fill_value="extrapolate")
        local_pfuzzines = approximate_tempo(alignment[:,0])*pfuzziness

    for i in range(len(alignment)-window_size):
        # extract all score notes with onsets inside the closed inter beat interval
        score_window_mask = np.all([score_note_array["onset_beat"] >= alignment[i,0]-sfuzziness,
                                    score_note_array["onset_beat"] < alignment[i+window_size,0]+sfuzziness], axis = 0)
        # extract all performance notes with onsets inside the inter beat interval plus some fuzzy relaxation
        performance_window_mask = np.all([performance_note_array["onset_sec"] >= alignment[i,1]-local_pfuzzines[i],
                                    performance_note_array["onset_sec"] < alignment[i+window_size,1]+local_pfuzzines[i]], axis = 0)
        score_note_arrays.append(score_note_array[score_window_mask])
        performance_note_arrays.append(performance_note_array[performance_window_mask])

    return score_note_arrays, performance_note_arrays


################################### SEGMENT MENDING ###################################


def mend_note_alignments(note_alignments, 
                 performance_note_array, 
                 score_note_array, 
                 node_times, 
                 symbolic_note_matcher, 
                 max_traversal_depth = 150):
    """
    mend note alignments in (overlapping) windows.
    creates a global dictionary of MAPS style alignments
    from a list of windowed MAPS style alignments
    """
                 
    score_alignment = {"insertion":[]}
    performance_alignment = {"deletion":[]}
    alignment = []

    # approximate_position = interp1d(node_times[:,0],node_times[:,1],fill_value="extrapolate")
    # approximate_tempo = interp1d(node_times[:-1,0],np.diff(node_times[:,1])/np.diff(node_times[:,0]),fill_value="extrapolate")

    # score_onsets_by_id = { snote["id"]:snote["onset_beat"] for snote in score_note_array}

    # combine all note alignments in two dictionaries
    for window_id in range(len(note_alignments)):
        for alignment_line in note_alignments[window_id]:
            if alignment_line["label"] == "match":
                if alignment_line["score_id"] in score_alignment.keys():
                    score_alignment[alignment_line["score_id"]].append(alignment_line["performance_id"])
                else:
                    score_alignment[alignment_line["score_id"]] = [alignment_line["performance_id"]]

                if alignment_line["performance_id"] in performance_alignment.keys():
                    performance_alignment[alignment_line["performance_id"]].append(alignment_line["score_id"])
                else:
                    performance_alignment[alignment_line["performance_id"]] = [alignment_line["score_id"]]

            if alignment_line["label"] == "deletion":
                if alignment_line["score_id"] in score_alignment.keys():
                    score_alignment[alignment_line["score_id"]].append("deletion")
                else:
                    score_alignment[alignment_line["score_id"]] = ["deletion"]

            if alignment_line["label"] == "insertion":
                if alignment_line["performance_id"] in performance_alignment.keys():
                    performance_alignment[alignment_line["performance_id"]].append("insertion")
                else:
                    performance_alignment[alignment_line["performance_id"]] = ["insertion"]

    # compute as single unique ids alignment
    used_perf_notes = ["deletion"]
    used_score_notes = ["insertion"]
    for snote_id in score_alignment.keys():

        # IF 1 p_note in alignment:
        if snote_id not in used_score_notes:
            if len(score_alignment[snote_id]) == 1:
                # DELETION
                if score_alignment[snote_id][0] == "deletion":
                    alignment.append({'label': 'deletion', 'score_id': snote_id})
                    used_score_notes.append(snote_id)
                else:
                    score_ids = []
                    performance_ids = []     
                    counter = 0
                    traverse_the_alignment_graph(snote_id, score_ids, performance_ids, performance_alignment, score_alignment, counter, max_depth=max_traversal_depth)
                    # deal with the possible alignments
                    no_insertion = [sid for sid in score_ids if sid != "insertion"]
                    no_deletion = [pid for pid in performance_ids if pid != "deletion"]
                    # if two unique notes match
                    if len(no_deletion) == 1 and len(no_insertion) == 1:
                        alignment.append({'label': 'match', 'score_id': snote_id, 'performance_id': str(score_alignment[snote_id][0])})
                        used_score_notes.append(no_insertion[0])
                        used_perf_notes.append(str(no_deletion[0]))
                    # try realigning
                    else:
                        local_score_note_array = score_note_array[np.any([score_note_array["id"] == str(idx) for idx in no_insertion], axis=0)]
                        masks = [performance_note_array["id"] == str(idx) for idx in no_deletion]
                        local_performance_note_array = performance_note_array[np.any(masks, axis=0)]
                        #print( snote_id, "realigning with score and performance array: ",local_score_note_array.shape, local_performance_note_array.shape)
                        
                        local_alignment = symbolic_note_matcher(local_score_note_array, local_performance_note_array, node_times)
                        #import pdb; pdb.set_trace()
                        used_score_notes += no_insertion
                        used_perf_notes += no_deletion
                        alignment += local_alignment

            
            else:
                # multiple perf_ids aligned to snote_id
                score_ids = []
                performance_ids = []     
                counter = 0
                traverse_the_alignment_graph(snote_id, score_ids, performance_ids, performance_alignment, score_alignment, counter, max_depth=max_traversal_depth)

                # deal with the possible alignments
                no_insertion = [sid for sid in score_ids if sid != "insertion"]
                no_deletion = [pid for pid in performance_ids if pid != "deletion"]
                # if two unique notes match
                if len(no_deletion) == 1 and len(no_insertion) == 1:
                    alignment.append({'label': 'match', 'score_id': snote_id, 'performance_id': str(no_deletion[0])})
                    used_score_notes.append(no_insertion[0])
                    used_perf_notes.append(str(no_deletion[0]))
                
                elif len(no_deletion) == 0:
                    # all deletions
                    alignment.append({'label': 'deletion', 'score_id': snote_id})
                    used_score_notes.append(snote_id)
                else:
                    # try realigning
                    
                    
                    local_score_note_array = score_note_array[np.any([score_note_array["id"] == str(idx) for idx in no_insertion], axis=0)]
                    local_performance_note_array = performance_note_array[np.any([performance_note_array["id"] == str(idx) for idx in no_deletion], axis=0)]
                    
                    #print( snote_id,"realigning (conflicting) with score and performance array: ",local_score_note_array.shape, local_performance_note_array.shape)
                    local_alignment = symbolic_note_matcher(local_score_note_array, local_performance_note_array, node_times)

                    #import pdb; pdb.set_trace()
                    used_score_notes += no_insertion
                    used_perf_notes += no_deletion
                    alignment += local_alignment


    for pnote_id in performance_alignment.keys():
        if pnote_id not in used_perf_notes:
            #print("previously unused performance note: ", pnote_id, performance_alignment[pnote_id])
            alignment.append({'label': 'insertion', 'performance_id': pnote_id})

    return alignment, score_alignment, performance_alignment





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
from setuptools import find_packages, setup


# Package meta-data.
NAME = 'parangonar'
DESCRIPTION = 'Symbolic music alignment'
KEYWORDS = 'match alignment midi performance score'
URL = "https://github.com/sildater/parangonar"
AUTHOR = 'Silvan Peter, Carlos Cancino-Chacón, Florian Henkel'
REQUIRES_PYTHON = '>=3.7'
VERSION = '0.0.10'

# What packages are required for this module to be executed?
REQUIRED = [
    'numpy',
    'scipy',
    'partitura>=1.1.0',
    'python-hiddenmarkov'
]

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION

# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=KEYWORDS,
    author=AUTHOR,
    url=URL,
    python_requires=REQUIRES_PYTHON,
    packages=find_packages(exclude=('tests',)),
        package_data={
        "parangonar": [
            "assets/mozart_k265_var1.match",
        ]
        },
    install_requires=REQUIRED,
    extras_require={},
    include_package_data=True,
    license="Apache 2.0",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module
evaluate note alignments symbolic music data.
"""
from typing import List
import numpy as np
import partitura as pt
import matplotlib.pyplot as plt
import random

def fscore_alignments(prediction: List[dict], 
                        ground_truth: List[dict], 
                        types: List[str]) -> (float, float, float):
    """
    Parameters
    ----------
    prediction: List of dictionaries containing the predicted alignments
    ground_truth: List of dictionaries containing the ground truth alignments
    types: List of alignment types to consider for evaluation (e.g ['match', 'deletion', 'insertion']

    Returns
    -------
    precision, recall, f score
    """

    pred_filtered = list(filter(lambda x: x['label'] in types, prediction))
    gt_filtered = list(filter(lambda x: x['label'] in types, ground_truth))

    filtered_correct = [pred for pred in pred_filtered if pred in gt_filtered]

    n_pred_filtered = len(pred_filtered)
    n_gt_filtered = len(gt_filtered)
    n_correct = len(filtered_correct)

    if n_pred_filtered > 0 or n_gt_filtered > 0:
        precision = n_correct / n_pred_filtered if n_pred_filtered > 0 else 0.
        recall = n_correct / n_gt_filtered if n_gt_filtered > 0 else 0
        f_score = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0
    else:
        # no prediction and no ground truth for a given type -> correct alignment
        precision, recall, f_score = 1., 1., 1.

    return precision, recall, f_score




def plot_alignment(ppart_na, 
         part_na, 
         alignment, 
         save_file = False,
         fname = "note_alignment",
         random_color = False
         ):
    
    first_note_midi = np.min(ppart_na["onset_sec"])
    last_note_midi = np.max(ppart_na["onset_sec"]+ppart_na["duration_sec"])
    first_note_start = np.min(part_na["onset_beat"])
    last_note_start = np.max(part_na["onset_beat"])
    length_of_midi = last_note_midi - first_note_midi
    length_of_xml = last_note_start - first_note_start

    length_of_pianorolls = max(10000,int(length_of_xml*8))
    time_div_midi = int(np.floor(length_of_pianorolls/length_of_midi))
    time_div_xml = int(np.floor(length_of_pianorolls/length_of_xml))
    
    midi_piano_roll, perfidx = pt.utils.compute_pianoroll(ppart_na,
                                                time_unit = "sec",
                                                time_div = time_div_midi,
                                                return_idxs=True,
                                                remove_drums=False)
    xml_piano_roll, scoreidx = pt.utils.compute_pianoroll(part_na,
                                                time_unit = "beat",
                                                time_div = time_div_xml,
                                                return_idxs = True,
                                                remove_drums=False)
    
    plot_array = np.zeros((128*2+50, length_of_pianorolls+800))
    dense_midi_pr = midi_piano_roll.todense()#[:,:time_size]
    dense_midi_pr[dense_midi_pr>0] = 1
    plot_array[:128,:xml_piano_roll.shape[1]] = xml_piano_roll.todense()
    plot_array[50+128:,:midi_piano_roll.shape[1]] = dense_midi_pr

    f, axs = plt.subplots(1,1,figsize=(100, 10))
    axs.matshow(plot_array, aspect = "auto",  origin='lower')
    hexadecimal_alphabets = '0123456789ABCDEF'

    if random_color:
        colors = ["#" + ''.join([random.choice(hexadecimal_alphabets) for j in range(6)]) for i in range(40)]
    else:
        colors = ["#00FF00" for i in range(40)]
        
    score_dict = dict()
    for note, pos in zip(part_na, scoreidx):
        score_dict[note["id"]] = pos
    perf_dict = dict()
    for note, pos in zip(ppart_na, perfidx):
        perf_dict[note["id"]] = pos
    for i, line in enumerate(alignment):
        if line["label"]=="match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot([score_pos[1], perf_pos[1]], [ score_pos[0], 128+50 +perf_pos[0]],'o-', lw=2, c = colors[i%40])

    if save_file:
        plt.savefig(fname+".png")
        plt.close(f)
    else:
        plt.show()





# def transform_labels(ids: List[str], 
#                      prediction: List[dict],
#                      ground_truth: List[dict], 
#                      field: str) -> (List[str], List[str]):
#     """
#     Parameters
#     ----------
#     ids: List of note ids
#     prediction: List of dictionaries containing the predicted alignments
#     ground_truth: List of dictionaries containing the ground truth alignments
#     field: which note ids to evaluate (score_id or performance_id)

#     Returns
#     -------
#     List of predicted labels, List of ground truth labels
#     """

#     assert field in ['score_id', 'performance_id']

#     y_pred, y_gt = [], []
#     for id in ids:

#         pred = list(filter(lambda x: field in x and str(x[field]) == id, prediction))
#         gt = list(filter(lambda x: field in x and str(x[field]) == id, ground_truth))
#         try:
#             assert len(pred) == len(gt) == 1
#             pred = pred[0]

#         except:
#             print(f'Missing note in predictions {id}')

#             if field == 'score_id':
#                 pred = dict(label='deletion',
#                             score_id=id)
#             elif field == 'performance_id':
#                 pred = dict(label='insertion',
#                             performance_id=id)
#             # import pdb
#             # pdb.set_trace()
            
#         gt = gt[0]

#         gt_label = gt['label']
#         pred_label = pred['label']

#         if gt_label == pred_label == 'match':
#             # matched to different note in performance -> mismatch
#             if pred['performance_id'] != gt['performance_id']:
#                 pred_label = 'mismatch'

#         y_gt.append(gt_label)
#         y_pred.append(pred_label)

#     return y_pred, y_gt









# def alignment_confusion_matrices(score_ids: List[str],
#                                  performance_ids: List[str],
#                                  prediction: List[dict],
#                                  ground_truth: List[dict],
#                                  normalize: str or None=None) -> (np.ndarray, np.ndarray):
#     """

#     Parameters
#     ----------
#     score_ids: List of score note ids
#     performance_ids: List of performance note ids
#     prediction: List of dictionaries containing the predicted alignments
#     ground_truth: List of dictionaries containing the ground truth alignments

#     Returns
#     -------
#     2 3x3 numpy array containing the confusion matrices for score and performance
#     (note that the last row (mismatch) should always be 0 for each column since does not occur in the ground truth
#     """
#     if normalize is not None:
#         assert normalize in ['true', 'pred', 'all']

#     y_pred_score, y_gt_score = transform_labels(score_ids, prediction, ground_truth, field='score_id')
#     y_pred_perf, y_gt_perf = transform_labels(performance_ids, prediction, ground_truth, field='performance_id')

#     conf_matrix_score = confusion_matrix(y_gt_score, y_pred_score,
#                                          labels=['match', 'deletion',  'mismatch'],
#                                          normalize=normalize)
#     conf_matrix_perf = confusion_matrix(y_gt_perf, y_pred_perf,
#                                         labels=['match', 'insertion', 'mismatch'],
#                                         normalize=normalize)

#     return conf_matrix_score, conf_matrix_perf